#!//usr/bin/perl
#
# model file converter
# (c) astx (at) aws-it.at, 2018
# and (c) others - see DIYaudio thread https://www.diyaudio.com/forums/software-tools/331656-installing-using-ngspice-opensource-simulator.html
#
# $Id: convert_sgp_modelfile_to_vbic.pl 1601 2019-02-22 17:38:47Z sta $
#
# (c) of sgp2vbic part derived from:
#
#	sgp2vbic: program to convert an SGP .model card to VBIC
#
# Vers	Date		Who		Comments
# ====	==========	=============	========
#  2.0	06/01/00	Colin McAndrew	translated from shell/nawk version
#

use strict;

my $debug=0;
my $verbose=0;
my $Number='([+-]?[0-9]+[.]?[0-9]*|[+-]?[0-9]+[.]?[0-9]*[eE][+-]?[0-9]+|[+-]?[.][0-9]+|[+-]?[.][0-9]+[eE][+-]?[0-9]+)';
my $Pi=atan2(0,-1);
my $Vbeif=0.6;
my $Vceif=2.5;
my $Vbcir=0.5;
my $Vecir=2.5;
my $outfile=undef;

my @order;
&load_output_order();

for (;;) {
	if ($#ARGV < 0) {
		last;
	} elsif ($ARGV[0] =~ /^-d/i) {
		$debug=1;
		$verbose=1;
	} elsif ($ARGV[0] =~ /^-h/i) {
		&usage();exit(0);
	} elsif ($ARGV[0] =~ /^-i/i) {
		&info();exit(0);
	} elsif ($ARGV[0] =~ /^-v$/i) {
		$verbose=1;
	} elsif ($ARGV[0] =~ /^-vbeif$/i) {
		shift(@ARGV);
		if (!defined($ARGV[0])) {
			die("ERROR: no value specified for -vbeif flag, stopped");
		}
		$Vbeif=$ARGV[0];
		if ($Vbeif !~ /$Number/) {
			die("ERROR: value for -vbeif flag must be a number, stopped");
		}
	} elsif ($ARGV[0] =~ /^-vceif$/i) {
		shift(@ARGV);
		if (!defined($ARGV[0])) {
			die("ERROR: no value specified for -vceif flag, stopped");
		}
		$Vceif=$ARGV[0];
		if ($Vceif !~ /$Number/) {
			die("ERROR: value for -vceif flag must be a number, stopped");
		}
	} elsif ($ARGV[0] =~ /^-vbcir$/i) {
		shift(@ARGV);
		if (!defined($ARGV[0])) {
			die("ERROR: no value specified for -vbcir flag, stopped");
		}
		$Vbcir=$ARGV[0];
		if ($Vbcir !~ /$Number/) {
			die("ERROR: value for -vbcir flag must be a number, stopped");
		}
	} elsif ($ARGV[0] =~ /^-vecir$/i) {
		shift(@ARGV);
		if (!defined($ARGV[0])) {
			die("ERROR: no value specified for -vecir flag, stopped");
		}
		$Vecir=$ARGV[0];
		if ($Vecir !~ /$Number/) {
			die("ERROR: value for -vecir flag must be a number, stopped");
		}
	} elsif ($ARGV[0] =~ /^-o$/i) {
		shift(@ARGV);
		if (!defined($ARGV[0])) {
			die("ERROR: no value specified for -o flag, stopped");
		}
		$outfile=$ARGV[0];
	} elsif ($ARGV[0] =~ /^-/) {
		&usage();
		die("ERROR: unknown flag $ARGV[0], stopped");
	} else {
		last;
	}
	shift(@ARGV);
}
if ($#ARGV < 0) {
	&usage();exit(1);	# exit if no file name is specified
}

my $ModelFile=$ARGV[0];

my @output;
open(INP, "< $ModelFile") && do {
	my @lines=<INP>;
	chomp(@lines);
	close(INP);
	my $i=0;
	my $scan_model=0;
	my $model='';
	for ($i=0;$i < @lines; $i++) {
		my $line=$lines[$i];
		$line=~s/\r//g;
		# print "$i=>$line\n";
		# normalize .model lines which are spread over many lines
		if ($line=~/^(\*|\;)/) {
			push(@output, $line);
			next;
		}		
		if ($line=~/^\s*$/) {
			push(@output, $line);
			next;
		}		

		# skip AKO lines: .model BC550C_Ckf AKO:BC550C_C
		if ($line=~/^(.model)\s+([^\s]+)\s+(ako)/i) {
			if ($verbose || $debug) {
				print STDERR "skip ako line: $line\n";
			}
			push(@output, $line);
			next;
		}
		if ($line=~/^.model/i) {
			# model parameter continuation on next line?
			if (($i + 1) < @lines && $lines[$i+1]=~/^\+/) {
				$scan_model=1;
				$model=$line;
			} else {
				$scan_model=0;
				# write model card
				push(@output, &convert_model_line($line));
			}
		} elsif ($scan_model && $line=~/^\+(.*)/) {
			$line=" $1";
			$model.=$line;
			# model parameter continuation on next line?
			if (($i + 1) < @lines && $lines[$i+1]=~/^\+/) {
				$scan_model=1;
			} else {
				$scan_model=0;
				# write model card
				push(@output, &convert_model_line($model));
				$model='';
			}
		} else {
			push(@output, $line);
		}
	}
};
if (defined $outfile && $outfile ne '') {
	open(OUT, "> $outfile") && do {
		print OUT join("\n", @output);
		close(OUT);
	};
} else {
	print join("\n", @output);
}
exit;

#------------------------------------------------------------------------------
# normalize model line
sub clean_model_line
#------------------------------------------------------------------------------
{
	my $model=shift;
	$model=~s/\(/ \( /g;
	$model=~s/\)/ \) /g;
	$model=~s/\s+/ /g;
	# remove spaces between key/value pair
	$model=~s/\s*=\s*/=/g;

	# add () if missing
	# .MODEL LS3250A NPN
	if ($model=~/^(.model) ([^\s]+) ([^\s]+) ([^\(])(.*)/i) {
		$model="$1 $2 $3 ( $4$5 )";
	}
	$model=~s/\s+$//g;
	return $model;
}

#------------------------------------------------------------------------------
sub normalize_units
#------------------------------------------------------------------------------
{
	my $val=shift;
	if ($debug) {
		print STDERR "normalize units:\t$val\t=>\t";
	}
	if    ($val=~/([\+\-\.0-9]+)f$/i)	 {	$val=$1 * 1e-15;	}
	elsif ($val=~/([\+\-\.0-9]+)p$/i)	 {	$val=$1 * 1e-12;	}
	elsif ($val=~/([\+\-\.0-9]+)n$/i)	 {	$val=$1 * 1e-9;		}
	elsif ($val=~/([\+\-\.0-9]+)u$/i)	 {	$val=$1 * 1e-6;		}
	elsif ($val=~/([\+\-\.0-9]+)m$/i)	 {	$val=$1 * 1e-3;		}
	elsif ($val=~/([\+\-\.0-9]+)k$/i)	 {	$val=$1 * 1e+3;		}
	elsif ($val=~/([\+\-\.0-9]+)meg$/i)	 {	$val=$1 * 1e+6;		}
	elsif ($val=~/([\+\-\.0-9]+)g$/i)	 {	$val=$1 * 1e+9;		}
	elsif ($val=~/([\+\-\.0-9]+)t$/i)	 {	$val=$1 * 1e+12;	}

	if ($debug) {
		print STDERR "$val\n";
	}
	return $val;
}

#------------------------------------------------------------------------------
# normalize model line
sub convert_model_line
#------------------------------------------------------------------------------
{
	my $model=shift;
	$model=&clean_model_line($model);
	if ($model=~/^(.model) ([^\s]+) (npn|pnp) \( (.*) \)$/i) {
		my $tag=$1;
		my $device=$2;
		my $type=$3;
		my $params=$4;
		my @params=split(/\s+/, $params);
		my $vals={};
		foreach my $line (@params) {
			my ($key, $value)=split(/=/, $line);
			$key=lc($key);
			# todo: normalize/convert here value to e00 notation
			$vals->{$key}=&normalize_units($value);
		}
		if (defined $vals->{'level'} && ($vals->{'level'} == 4 || $vals->{'level'} == 9)) {
			if ($verbose) {
				print STDERR "skip sgp2vbic conversion for already converted model $device\n";
			}
		} else {
			$model=&convert_sgp_to_vbic($device, $type, $vals);
		}
	}
	return $model;
}

#------------------------------------------------------------------------------
sub QCDEPL
#------------------------------------------------------------------------------
{
	my $vj=shift;
	my $p=shift;
	my $m=shift;
	my $f=shift;

        my $w;
	my $xx;
	my $cj;
	my $qj;

	if ($debug) {
		print STDERR "QCDEPL: $vj, $p, $m, $f\n";
	}

	$w=1.0 - $vj / $p;
	if ($w >= (1.0 - $f)){
		$cj=$w ** (-$m);
		$qj=$p * (1.0 - $w * $cj) / (1.0 - $m);
	} else {
		$xx=(1.0 - $f) ** (-(1.0 + $m));
		$cj=$xx * (1.0 - $f * (1.0 + $m) + $m * $vj / $p);
		$qj=$xx * ((1.0 - $f * (1.0 + $m)) * ($vj - $f * $p) + 0.5 * $m * ($vj * $vj - $f * $f * $p * $p) / $p) + $p * (1.0 - $xx * (1.0 - $f) ** 2) / (1.0 - $m);
	}

	if ($debug) {
		print STDERR "QCDEPL: result: $qj, $cj\n";
	}

	return ($qj, $cj);
}

#------------------------------------------------------------------------------
sub load_output_order
#------------------------------------------------------------------------------
{
	@order=(
		'level',
		'rcx',
		'rci',
		'gamm',
		'vo',
		'qco',
		'rbx',
		'rbi',
		're',
		'is',
		'nf',
		'nr',
		'fc',
		'cje',
		'pe',
		'me',
		'cjc',
		'cjep',
		'pc',
		'mc',
		'cjcp',
		'ps',
		'ms',
		'ibei',
		'nei',
		'iben',
		'nen',
		'ibci',
		'nci',
		'ibcn',
		'ncn',
		'vef',
		'ver',
		'ikf',
		'nkf',		# Supported in LTspice VBIC v1.2
		'ikr',
		'rci',
		'tf',
		'xtf',
		'vtf',
		'itf',
		'tr',
		'td',
		'ea',
		'eaie',
		'eaic',
		'eane',
		'eanc',
		'xis',
		'xii',
		'xin',
		'kfn',
		'afn',
	);
}

#------------------------------------------------------------------------------
sub convert_sgp_to_vbic
#------------------------------------------------------------------------------
{
	my $name=shift;
	my $type=shift;
	my $ref=shift;

	my $model='';
	my $val={};
	$val->{'level'}=9;		# VBIC is level 4 or level 9 (ltspice?) npn/pnp
	$val->{'ibe'}=undef;		# 
	$val->{'ibc'}=undef;		#
	$val->{'is'}=1.0e-16;
	$val->{'bf'}=100.0;
	$val->{'nf'}=1.0;
	$val->{'vaf'}=0.0;
	$val->{'ikf'}=0.0;
        $val->{'nk'}=0.5;		# Supported by LTspice modified SGP and VBIC v1.2
	$val->{'ise'}=0.0;
	$val->{'ne'}=1.5;
	$val->{'br'}=1.0;
	$val->{'nr'}=1.0;
	$val->{'var'}=0.0;
	$val->{'ikr'}=0.0;
	$val->{'isc'}=0.0;
	$val->{'nc'}=2.0;
	$val->{'rb'}=0.0;
	$val->{'rbm'}=0.0;
	$val->{'re'}=0.0;
	$val->{'rc'}=0.0;
        $val->{'rco'}=0.0;		# Supported by LTspice modified SGP and VBIC v1.2
        $val->{'qco'}=0.0;		# ^
        $val->{'gamma'}=1.0e-11;	# ^
        $val->{'vo'}=0.0;			# ^
	$val->{'cje'}=0.0;
	$val->{'vje'}=0.75;
	$val->{'mje'}=0.33;
	$val->{'cjc'}=0.0;
	$val->{'vjc'}=0.75;
	$val->{'mjc'}=0.33;
	$val->{'xcjc'}=1.0;
	$val->{'cjs'}=0.0;
	$val->{'vjs'}=0.75;
	$val->{'mjs'}=0.0;
	$val->{'fc'}=0.5;
	$val->{'tf'}=0.0;
	$val->{'xtf'}=0.0;
	$val->{'vtf'}=0.0;
	$val->{'itf'}=0.0;
	$val->{'ptf'}=0.0;
	$val->{'tr'}=0.0;
	$val->{'kf'}=0.0;
	$val->{'af'}=1.0;
	$val->{'eg'}=1.11;
	$val->{'xtb'}=0.0;
	$val->{'xti'}=3.0;

	my $alias={};
	$alias->{'va'}='vaf';
	$alias->{'ik'}='ikf';
	$alias->{'c2'}='ise';
	$alias->{'vb'}='var';
	$alias->{'c4'}='isc';
	$alias->{'pe'}='vje';
	$alias->{'me'}='mje';
	$alias->{'pc'}='vjc';
	$alias->{'mc'}='mjc';
	$alias->{'ccs'}='cjs';
	$alias->{'ps'}='vjs';
	$alias->{'ms'}='mjs';
	$alias->{'pt'}='xti';

	my $ignore={};
	$ignore->{'mfg'}=1;

	# always keep some info values as comment
	my $keep={};
	$keep->{'mfg'}=1;
	$keep->{'vceo'}=1;
	$keep->{'icrating'}=1;
 
	my $done={};
	my $r={};
	my $info={};

	foreach my $key (sort keys %{$ref}) {
		my $value=$ref->{$key};
		if ($debug) {
			print STDERR "device $name $type: $key=$value\n";
		}
		if (!defined $ignore->{$key} && $value !~ /$Number/) {
			print STDERR "*** ERROR - device $name $type: parameter $key must be a number, skipped device\n";
			return;
		}
		if (defined($alias->{$key})) {
			$key=$alias->{$key};
		}
		if ($key eq 'irb') {
			print STDERR "=== WARNING - device $name $type: IRB parameter is not supported in vbic\n";
		}
		if (!defined($val->{$key})) {
			print STDERR "=== WARNING - device $name $type: parameter $key is not supported in vbic\n";
		}
		if (defined $keep->{$key}) {
			# will become a comment only!
			$info->{$key}=$value;
		} else {
			$val->{$key}=$value;
		}
		if ($key eq 'rbm') {$done->{'rbm'}='yes';}
		if ($key eq 'ibe') {$done->{'ibe'}='yes';}
		if ($key eq 'ibc') {$done->{'ibc'}='yes';}
	}
	if (!defined($done->{'rbm'})) {
		$val->{'rbm'}=$val->{'rb'};
	}
	if (!defined($done->{'ibe'})) {		# Prefer ibe to is, if it is available
		$val->{'ibe'}=$val->{'is'};
	}
	if (!defined($done->{'ibc'})) {		# Prefer ibc to is, if it is available
		$val->{'ibc'}=$val->{'is'};
	}
	if ($val->{'ise'} > 1.0) {
		$val->{'ise'}=$val->{'ise'} * $val->{'is'};
	}
	if ($val->{'isc'} > 1.0) {
		$val->{'isc'}=$val->{'isc'} * $val->{'is'};
	}
	if ($val->{'mjc'} == 1.0) {
		print STDERR "=== WARNING - device $name $type: parameter mjc must not be 1.0, would raise a division by zero in function QCDEPL. Setting default value to 0.33\n";
		$val->{'mjc'}=0.33;
	}
	if ($val->{'mje'} == 1.0) {
		print STDERR "=== WARNING - device $name $type: parameter mje must not be 1.0, would raise a division by zero in function QCDEPL. Setting default value to 0.33\n";
		$val->{'mje'}=0.33;
	}

	my $Vbcif=$Vbeif - $Vceif;
	my $Vbeir=$Vbcir - $Vecir;
	my ($qjbef, $cj   )=&QCDEPL($Vbeif, $val->{'vje'}, $val->{'mje'}, $val->{'fc'});
	my ($qjbcf, $cjbcf)=&QCDEPL($Vbcif, $val->{'vjc'}, $val->{'mjc'}, $val->{'fc'});
	my ($qjber, $cjber)=&QCDEPL($Vbeir, $val->{'vje'}, $val->{'mje'}, $val->{'fc'});
	my ($qjbcr, $cj   )=&QCDEPL($Vbcir, $val->{'vjc'}, $val->{'mjc'}, $val->{'fc'});
	my $ivaf=$val->{'vaf'};
	if ($ivaf > 0) {
		$ivaf=1/$ivaf;
	}
	my $ivar=$val->{'var'};
	if ($ivar > 0) {
		$ivar=1/$ivar;
	}
	my $godIf=$ivaf / (1 - $Vbeif * $ivar - $Vbcif * $ivaf);
	if ($godIf < 1e-10) {
		$godIf=1e-10;
	}
	my $godIr=$ivar / (1 - $Vbeir * $ivar - $Vbcir * $ivaf);
	if ($godIr < 1e-10) {
		$godIr=1e-10;
	}
	my $a11=$qjbcf-$cjbcf/$godIf;
	my $a12=$qjbef;
	my $r1=-1.0;
	my $a21=$qjbcr;
	my $a22=$qjber-$cjber/$godIr;
	my $r2=-1.0;
	my $det=$a11 * $a22 - $a12 * $a21;
	my $ivef=($r1 * $a22 - $r2 * $a12) / $det;
	my $iver=($r2 * $a11 - $r1 * $a21) / $det;
	my $vef=1 / $ivef;
	my $ver=1 / $iver;


	$r->{'rcx'}	=	$val->{'rc'};
	$r->{'rci'}	=	$val->{'rco'};			# Supported by LTspice modified SGP and VBIC v1.2
        $r->{'qco'}     =       $val->{'qco'};			# ^
        $r->{'gamm'}    =       $val->{'gamma'};		# ^
        $r->{'vo'}      =       $val->{'vo'};			# ^
#	$r->{'qbm'}	=	'1.0';				# Enable SGP formulation with irb
	$r->{'rbx'}	=	$val->{'rbm'};
	$r->{'rbi'}	=	($val->{'rb'} - $val->{'rbm'});
	$r->{'re'}	=	$val->{'re'};
	$r->{'is'}	=	$val->{'ibe'};			# Uses Is if not available
	$r->{'nf'}	=	$val->{'nf'};
	$r->{'nr'}	=	$val->{'nr'};
	$r->{'fc'}	=	$val->{'fc'};
	$r->{'cje'}	=	$val->{'cje'};
	$r->{'pe'}	=	$val->{'vje'};
	$r->{'me'}	=	$val->{'mje'};
	$r->{'cjc'}	=	($val->{'cjc'} * $val->{'xcjc'});
	$r->{'cjep'}	=	($val->{'cjc'} * (1.0 - $val->{'xcjc'}));
	$r->{'pc'}	=	$val->{'vjc'};
	$r->{'mc'}	=	$val->{'mjc'};
	$r->{'cjcp'}	=	$val->{'cjs'};
	$r->{'ps'}	=	$val->{'vjs'};
	$r->{'ms'}	=	$val->{'mjs'};
	$r->{'ibei'}	=	($val->{'ibe'} / $val->{'bf'});	# Use ibe if it is available
	$r->{'nei'}	=	$val->{'nf'};
	$r->{'iben'}	=	$val->{'ise'};
	$r->{'nen'}	=	$val->{'ne'};
	$r->{'ibci'}	=	($val->{'ibc'} / $val->{'br'}); # Use ibc if it is available
	$r->{'nci'}	=	$val->{'nr'};
	$r->{'ibcn'}	=	$val->{'isc'};
	$r->{'ncn'}	=	$val->{'nc'};
	$r->{'vef'}	=	$vef;
	$r->{'ver'}	=	$ver;
	$r->{'ikf'}	=	$val->{'ikf'};
        $r->{'nkf'}     =       $val->{'nk'};			# Supported by LTspice modified SGP and VBIC v1.2
	$r->{'ikr'}	=	$val->{'ikr'};
	$r->{'tf'}	=	$val->{'tf'};
	$r->{'xtf'}	=	$val->{'xtf'};
	$r->{'vtf'}	=	$val->{'vtf'};
	$r->{'itf'}	=	$val->{'itf'};
	$r->{'tr'}	=	$val->{'tr'};
	$r->{'td'}	=	($val->{'tf'} * $val->{'ptf'} * $Pi / 180.0);
	$r->{'ea'}	=	$val->{'eg'};
	$r->{'eaie'}	=	$val->{'eg'};
	$r->{'eaic'}	=	$val->{'eg'};
	$r->{'eane'}	=	$val->{'eg'};
	$r->{'eanc'}	=	$val->{'eg'};
	$r->{'xis'}	=	$val->{'xti'};

	$r->{'xii'}	=	($val->{'xti'} - $val->{'xtb'});
	# $r->{'xin'}	=	($val->{'xti'} - $val->{'ne'} * $val->{'xtb'});
	$r->{'xin'}	=	($val->{'xti'} - $val->{'xtb'});	# corrected according to the url: http://www.designers-guide.org/VBIC/documents/VbicText.pdf
	$r->{'kfn'}	=	$val->{'kf'};
	$r->{'afn'}	=	$val->{'af'};

	# shorten calculated values
	foreach my $o (keys %{$r}) {
		if ($r->{$o} =~ /$Number/ ) {
			if (int($r->{$o}) != $r->{$o}) {
				my $snt=sprintf("%8e", $r->{$o});
				if ($snt =~ /e[+-]00$/i) {
					$r->{$o}=sprintf("%0.8f", $r->{$o});
				} else {
					$r->{$o}=$snt;
				}
			}
			if ($r->{$o} =~ /\./) {
				if ($r->{$o} =~ /e/i) {
					$r->{$o} =~ s/[0]+e/e/gi;
					$r->{$o} =~ s/\.e/.0e/gi;
				} else {
					$r->{$o} =~ s/[0]+$//g;
					$r->{$o} =~ s/\.$//g;
				}
			}
		}
	}

	# output comment values
	my @outvals=();
	$model="* " . '-' x 78 . "\n";
	foreach my $o (sort keys %{$info}) {
		push(@outvals, "$o=$info->{$o}");
	}
	$model.="* $name: " . join(' ', @outvals) . "\n";
	$model.="* " . '-' x 78 . "\n";
	$model.=".model $name $type ( $val->{level}\n";
	my $i=0;
	my $val_per_line=6;
	my @keys=sort keys %{$r};
	my $nk=@keys;
	@outvals=();
	foreach my $o (@order) {
		if (defined $r->{$o}) {
			push(@outvals, "$o=$r->{$o}");
			delete $r->{$o};
			$i++;
		}
		if ($i % $val_per_line == 0) {
			if (@outvals) {
				$model.="+ " . join(' ', @outvals) . "\n";
				@outvals=();
			}
		}
	}
	# if other parameters out of order are in there, add it on end
	foreach my $o (sort keys %{$r}) {
		push(@outvals, "$o=$r->{$o}");
		delete $r->{$o};
		$i++;
		if ($i % $val_per_line == 0) {
			$model.="+ " . join(' ', @outvals) . "\n";
			@outvals=();
		}
	}
	if (@outvals) {
		$model.="+ " . join(' ', @outvals) . "\n";
	}
	$model.="+ )\n";
	$model.="* " . '-' x 78 . "\n";

	return $model;
}

sub usage() {
	print "
$0: convert SGP .model card to VBIC .model card

Usage: $0 [options] modelFile

Files:
	modelFile	file with SGP .model card

Options:
	-d		debug mode
	-h		print this help message
	-i		print detailed information
	-o outfile	output to file instead of STDOUT
	-v		verbose mode
	-vbeif Vbe	do fwd Early voltage map at Vbe ($Vbeif)
	-vceif Vce	do fwd Early voltage map at Vce ($Vceif)
	-vbcir Vbc	do rev Early voltage map at Vbc ($Vbcir)
	-vecir Vec	do rev Early voltage map at Vec ($Vecir)
";
} # End of: sub usage

sub info() {
	print "
This program maps an SGP .model card into a VBIC .model card.
For many parameters there is a 1-to-1 mapping between the two,
and the default VBIC parameters are such that the model extensions
that are not part of SGP are turned off.

There are two fundamental and non-mappable differences between
the two models. First, the IRB emitter crowding based resistance
modulation model is not supported in VBIC. This parameter is
ignored. Second, the Early effect model is different, the bias
dependence is substantially better in VBIC than in SGP. This means
the models can be made to match only at specific biases.
These biases can be specified by the -vxxi[fr] flags.
";
} # End of: sub info
